﻿'The following code example demonstrates how to change the behavior of the ENTER key in a DataGridView subclass 
'by overriding the ProcessDataGridViewKey and ProcessDialogKey methods. In the example, the ENTER key has the 
'same behavior as the RIGHT ARROW key, making it easier for a user to edit multiple cells in a single row of data
'
Imports System
Imports System.Windows.Forms
Imports System.Security.Permissions

Public Class DataGridViewEnterKey

    Inherits DataGridView

    '<System.Security.Permissions.UIPermission(SecurityAction.LinkDemand, Window:=UIPermissionWindow.AllWindows)>
    Protected Overrides Function ProcessDialogKey(ByVal keyData As Keys) As Boolean
        ' Extract the key code from the key value. 
        Dim key As Keys = keyData And Keys.KeyCode
        ' Handle the ENTER key as if it were a RIGHT ARROW key. 
        If key = Keys.Enter Then
            Return Me.ProcessRightKey(keyData)
        End If
        Return MyBase.ProcessDialogKey(keyData)
    End Function

    <SecurityPermission(SecurityAction.LinkDemand, Flags:=SecurityPermissionFlag.UnmanagedCode)>
    Protected Overrides Function ProcessDataGridViewKey(ByVal e As KeyEventArgs) As Boolean
        ' Handle the ENTER key as if it were a RIGHT ARROW key. 
        If e.KeyCode = Keys.Enter Then
            Return Me.ProcessRightKey(e.KeyData)
        End If
        Return MyBase.ProcessDataGridViewKey(e)
    End Function

End Class