﻿Option Explicit On
Option Compare Binary
Option Strict On
Imports System.Math
Imports System.Double

Module ServiceSub

    Friend Sub ModifyingMatrixFactorization(ByRef Hij() As Double, ByRef pk() As Double, ByRef z() As Double, ByRef c As Double, ByVal eps0 As Double)
        'пересчитать U, D, вычислить u
        Dim i, j, jj As Integer
        Dim dj, bj, uj As Double
        jj = 0
        For j = 0 To n - 2
            dj = Hij(jj)
            uj = pk(j)
            Hij(jj) = Max(dj + c * (z(j) ^ 2), eps0)
            bj = z(j) * c / Hij(jj)
            c *= (dj / Hij(jj))
            For i = j + 1 To n - 1
                z(i) -= (z(j) * Hij(jj + i - j))
                Hij(jj + i - j) += (bj * z(i))
                pk(i) -= uj * Hij(jj + i - j)
            Next
            jj += n - j
        Next
        Hij(jj) = Max(Hij(jj) + (c * (z(n - 1) ^ 2)), eps0)
    End Sub

    Friend Sub p(ByRef Hij() As Double, ByRef pk() As Double)
        'построение направление спуска
        Dim i, j, k As Integer
        Dim r As Double
        For i = n - 1 To 1 Step -1
            r = pk(i)
            k = i
            For j = 0 To i - 1
                pk(j) -= r * Hij(k)
                k += n - 1 - j
            Next
        Next
    End Sub

    Friend Sub fib(ByVal nk As Integer, ByRef Fk1 As Double, ByRef xk1() As Double, ByRef pk() As Double, ByRef alfak As Double, ByVal hp As Double, ByVal a As Double, ByVal b As Double)
        'минимизация вдоль направления спуска
        '   min Fk1 = F(xk1 = xk + t * pk), a <= t <= b
        '   подход Хакимовой путем интеграции подходов Полака, Данилина и Пшеничного, Гилла и Мюррея к выбору величины шага
        '   величина шага t = b - a для проверки неравенства 
        '      Fk1 - Fk <= alfa * (b - a) * hp
        '      задается численно устойчивым методом Фибоначчи
        'nk - число итераций
        Dim i As Integer
        Dim kf1 As Long = 1
        Dim kf2 As Long = 2
        Dim kf As Long
        For i = 2 To nk
            kf = kf1 : kf1 = kf2 : kf2 = kf + kf1
        Next
        Dim ba As Double = (b - a) / kf2
        Dim z As Double = a + ba * kf1
        alfak = z : Call xk1i(n, alfak, xk, pk, xk1)
        Dim Fz As Double = func(sF, xk1, False)
        Fk1 = Fz : nf += 1
        If Fk1 - Fk <= alfa * z * hp Then Exit Sub
        Dim y As Double = a + ba * kf
        alfak = y : Call xk1i(n, alfak, xk, pk, xk1)
        Dim Fy As Double = func(sF, xk1, False)
        Fk1 = Fy : nf += 1
        If Fk1 - Fk <= alfa * y * hp Then Exit Sub
        For i = 2 To nk
            kf2 = kf1 : kf1 = kf : kf = kf2 - kf1
            If Fy > Fz Then
                a = y : y = z : Fy = Fz : z = a + ba * kf1
                alfak = z : Call xk1i(n, alfak, xk, pk, xk1)
                Fz = func(sF, xk1, False)
                Fk1 = Fz : nf += 1
            Else
                b = z : z = y : Fz = Fy : y = a + ba * kf
                alfak = y : Call xk1i(n, alfak, xk, pk, xk1)
                Fy = func(sF, xk1, False)
                Fk1 = Fy : nf += 1
            End If
            If Fk1 - Fk <= alfa * (b - a) * hp Then Exit For
        Next
    End Sub

    Friend Sub xk1i(ByVal n As Integer, ByVal alfak As Double, ByVal xk() As Double, ByVal pk() As Double, ByRef xk1() As Double)
        'xk1(i) = xk(i) + alfak * pk(i)
        Dim i As Integer
        For i = 0 To n - 1
            xk1(i) = xk(i) + alfak * pk(i)
        Next
    End Sub

    Friend Sub uUD(ByRef Hij() As Double, ByRef pk() As Double, ByVal epsh As Double, ByVal bhi As Boolean)
        'подход Хакимовой к построению u, U, D
        Dim i, j, ij As Integer
        Dim r, l As Double
        Dim kk As Integer = 0
        For k As Integer = 0 To n - 2
            If Hij(kk) < epsh Then
                l = -epsh
            Else
                l = Hij(kk)
            End If
            If Abs(Hij(kk)) < epsh Then
                'нулевое собственное значение
                n0 += 1
            Else
                If Hij(kk) < 0 Then
                    'отрицательное собственное значение
                    no += 1
                    If bhi Then
                        'седловая точка
                        ns += 1
                        For i = 0 To n - 1
                            pk(i) = 0
                        Next
                        pk(k) = 1
                        Exit Sub
                    End If
                End If
            End If
            Hij(kk) = Max(Hij(kk), epsh)
            Hij(kk) = Max(Hij(kk), Abs(pk(k)))
            For i = kk + 1 To kk + n - 1 - k
                Hij(kk) = Max(Hij(kk), Abs(Hij(i)))
            Next
            If l < 0 Then
                If Hij(kk) > epsh * 1.618 Then
                    l = Hij(kk) / 1.618
                Else
                    l = Abs(1)
                End If
            Else
                If Hij(kk) > l * 1.3 Then
                    Hij(kk) = l
                    l /= 1.3
                End If
            End If
            ij = kk - k
            For i = k + 1 To n - 1
                r = Hij(kk + i - k) / Hij(kk)
                ij += n - i
                For j = i To n - 1
                    Hij(ij + j) -= Hij(kk + j - k) * r
                Next
                pk(i) -= pk(k) * r
                Hij(kk + i - k) = r
            Next
            pk(k) *= Hij(kk) / (l * l)
            kk += n - k
        Next
        If Hij(kk) < epsh Then
            l = -epsh
        Else
            l = Hij(kk)
        End If
        Hij(kk) = Max(Hij(kk), epsh)
        Hij(kk) = Max(Hij(kk), Abs(pk(n - 1)))
        If l < 0 Then
            If Hij(kk) > epsh * 1.618 Then
                l = Hij(kk) / 1.618
            Else
                l = Abs(1)
            End If
        Else
            If Hij(kk) > l * 1.3 Then
                Hij(kk) = l
                l /= 1.3
            End If
        End If
        pk(n - 1) *= Hij(kk) / (l * l)
    End Sub

    Friend Sub Extrapolation(ByRef iex As Integer, ByRef ext(,) As Double, ByRef pk() As Double, ByRef bex As Boolean, ByRef nex As Integer)
        'подход Хакимовой к построению альтернативного направления спуска
        'ex(n - 1, 2): стартовые точки ex(i, iex), i = 0, ..., n - 1, iex = 0, 1, 2
        'iex - индекс стартовой точки ex(i, iex), i = 0, ..., n - 1
        'bex - признак вычисления альтернативного направления спуска
        'nex: -1 -> неоределенность, 0 -> Fk1 < Fk, 1 -> Fk1 > Fk
        Dim a, b, c, r21, r2110 As Double
        Dim i, j As Integer
        nex = -1
        For i = 0 To n - 1
            ext(i, iex) = xk(i)
        Next
        Select Case iex
            Case 3
                If bex Then
                    Dim berr As Boolean = False
                    For j = 0 To n - 1
                        r21 = ext(j, 2) - ext(j, 1)
                        r2110 = r21 / (ext(j, 1) - ext(j, 0))
                        If IsNaN(r2110) Or IsInfinity(b) Then
                            berr = True
                            Exit For
                        End If
                        c = ((ext(j, 3) - ext(j, 1)) / ((ext(j, 2) - ext(j, 0))) - r2110) / r21
                        If IsNaN(c) Or IsInfinity(b) Then
                            berr = True
                            Exit For
                        End If
                        b = r2110 - c * (ext(j, 1) + ext(j, 0))
                        a = ext(j, 1) - ext(j, 0) * (b + c * ext(j, 0))
                        pk(j) = (a + ext(j, 3) * (b + c * ext(j, 3)))
                    Next
                    If Not berr Then
                        nex = 1
                        na += 1
                        Call ChoiceSizeStep(pk, 0.8, 1.2, nex)
                        If nex = 0 Then nau += 1
                    End If
                End If
                For i = 0 To n - 1
                    ext(i, 0) = ext(i, 1)
                    ext(i, 1) = ext(i, 2)
                    ext(i, 2) = ext(i, 3)
                Next
            Case Else
                iex += 1
        End Select
    End Sub

    Friend Sub ChoiceSizeStep(ByRef pk() As Double, ByVal a As Double, ByVal b As Double, ByRef nex As Integer)
        'подход Хакимовой к минимизация вдоль альтернативного направления спуска
        'интеграция подходов Гольдштейна, Армийо, Полака
        Dim i, j As Integer
        Dim r, Fk1, tn, tw, hp As Double
        Dim x(n - 1), hi(n - 1) As Double
        If k = 4 Then
            For i = 0 To n - 1
                x(i) = 0
            Next
            Fk1 = func(sF, x, False)
            nf += 1
            If Fk1 < Fk Then
                nex = 0
                Fk = Fk1
                For i = 0 To n - 1
                    xk(i) = 0
                Next
                Exit Sub
            End If
        End If
        ApproximationDerivatives(hi, 0)
        For j = 1 To mex
            r = (a + b) * 0.5
            hp = 0
            For i = 0 To n - 1
                x(i) = pk(i) * r
                hp += hi(i) * (x(i) - xk(i))
            Next
            Fk1 = func(sF, x, False)
            nf += 1
            tw = Fk1 - Fk
            If tw < 0 Then
                nex = 0
                Fk = Fk1
                For i = 0 To n - 1
                    xk(i) = x(i)
                Next
            Else
                If j = 1 Then Exit For
            End If
            tn = tw - (1 - alfa) * hp
            tw -= alfa * hp
            If tn <= 0 Then
                a = r
            Else
                b = r
            End If
        Next
    End Sub

End Module
